"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("graphql");
const graphql = __importStar(require("graphql"));
const graphql_extensions_1 = require("graphql-extensions");
const apollo_cache_control_1 = require("apollo-cache-control");
const apollo_tracing_1 = require("apollo-tracing");
const apollo_server_errors_1 = require("apollo-server-errors");
const crypto_1 = require("crypto");
const requestPipelineAPI_1 = require("apollo-server-core/dist/requestPipelineAPI");
exports.InvalidGraphQLRequestError = requestPipelineAPI_1.InvalidGraphQLRequestError;
const dispatcher_1 = require("apollo-server-core/dist/utils/dispatcher");
// ADDED LINE //
const groupResolve_1 = __importDefault(require("../groupResolve"));
function computeQueryHash(query) {
    return crypto_1.createHash('sha256')
        .update(query)
        .digest('hex');
}
async function processGraphQLRequestNew(config, requestContext) {
    let cacheControlExtension;
    const extensionStack = initializeExtensionStack();
    requestContext.context._extensionStack = extensionStack;
    const dispatcher = initializeRequestListenerDispatcher();
    initializeDataSources();
    const request = requestContext.request;
    let { query, extensions } = request;
    let queryHash;
    let persistedQueryHit = false;
    let persistedQueryRegister = false;
    if (extensions && extensions.persistedQuery) {
        // It looks like we've received a persisted query. Check if we
        // support them.
        if (!config.persistedQueries || !config.persistedQueries.cache) {
            throw new apollo_server_errors_1.PersistedQueryNotSupportedError();
        }
        else if (extensions.persistedQuery.version !== 1) {
            throw new requestPipelineAPI_1.InvalidGraphQLRequestError('Unsupported persisted query version');
        }
        queryHash = extensions.persistedQuery.sha256Hash;
        if (query === undefined) {
            query = await config.persistedQueries.cache.get(`apq:${queryHash}`);
            if (query) {
                persistedQueryHit = true;
            }
            else {
                throw new apollo_server_errors_1.PersistedQueryNotFoundError();
            }
        }
        else {
            const computedQueryHash = computeQueryHash(query);
            if (queryHash !== computedQueryHash) {
                throw new requestPipelineAPI_1.InvalidGraphQLRequestError('provided sha does not match query');
            }
            persistedQueryRegister = true;
            // Store the query asynchronously so we don't block.
            (async () => {
                return config.persistedQueries && config.persistedQueries.cache.set(`apq:${queryHash}`, query);
            })().catch(error => {
                console.warn(error);
            });
        }
    }
    else if (query) {
        // FIXME: We'll compute the APQ query hash to use as our cache key for
        // now, but this should be replaced with the new operation ID algorithm.
        queryHash = computeQueryHash(query);
    }
    else {
        throw new requestPipelineAPI_1.InvalidGraphQLRequestError('Must provide query string.');
    }
    requestContext.queryHash = queryHash;
    const requestDidEnd = extensionStack.requestDidStart({
        request: request.http,
        queryString: request.query,
        operationName: request.operationName,
        variables: request.variables,
        extensions: request.extensions,
        persistedQueryHit,
        persistedQueryRegister,
        context: requestContext.context,
        requestContext,
    });
    const parsingDidEnd = await dispatcher.invokeDidStartHook('parsingDidStart', requestContext);
    try {
        let document;
        try {
            document = parse(query);
            parsingDidEnd();
        }
        catch (syntaxError) {
            parsingDidEnd(syntaxError);
            return sendResponse({
                errors: [
                    apollo_server_errors_1.fromGraphQLError(syntaxError, {
                        errorClass: apollo_server_errors_1.SyntaxError,
                    }),
                ],
            });
        }
        requestContext.document = document;
        const validationDidEnd = await dispatcher.invokeDidStartHook('validationDidStart', requestContext);
        const validationErrors = validate(document);
        if (validationErrors.length > 0) {
            validationDidEnd(validationErrors);
            return sendResponse({
                errors: validationErrors.map(validationError => apollo_server_errors_1.fromGraphQLError(validationError, {
                    errorClass: apollo_server_errors_1.ValidationError,
                })),
            });
        }
        validationDidEnd();
        // FIXME: If we want to guarantee an operation has been set when invoking
        // `willExecuteOperation` and executionDidStart`, we need to throw an
        // error here and not leave this to `buildExecutionContext` in
        // `graphql-js`.
        const operation = graphql_1.getOperationAST(document, request.operationName);
        requestContext.operation = operation || undefined;
        // We'll set `operationName` to `null` for anonymous operations.
        requestContext.operationName = (operation && operation.name && operation.name.value) || null;
        await dispatcher.invokeHookAsync('didResolveOperation', requestContext);
        const executionDidEnd = await dispatcher.invokeDidStartHook('executionDidStart', requestContext);
        let response;
        try {
            response = (await execute(document, request.operationName, request.variables));
            executionDidEnd();
        }
        catch (executionError) {
            executionDidEnd(executionError);
            return sendResponse({
                errors: [apollo_server_errors_1.fromGraphQLError(executionError)],
            });
        }
        const formattedExtensions = extensionStack.format();
        if (Object.keys(formattedExtensions).length > 0) {
            response.extensions = formattedExtensions;
        }
        if (config.formatResponse) {
            response = config.formatResponse(response, {
                context: requestContext.context,
            });
        }
        return sendResponse(response);
    }
    finally {
        requestDidEnd();
    }
    function parse(query) {
        const parsingDidEnd = extensionStack.parsingDidStart({
            queryString: query,
        });
        try {
            return graphql.parse(query);
        }
        finally {
            parsingDidEnd();
        }
    }
    function validate(document) {
        let rules = graphql_1.specifiedRules;
        if (config.validationRules) {
            rules = rules.concat(config.validationRules);
        }
        const validationDidEnd = extensionStack.validationDidStart();
        try {
            return graphql.validate(config.schema, document, rules);
        }
        finally {
            validationDidEnd();
        }
    }
    async function execute(document, operationName, variables) {
        const executionArgs = {
            schema: config.schema,
            document,
            rootValue: typeof config.rootValue === 'function' ? config.rootValue(document) : config.rootValue,
            contextValue: requestContext.context,
            variableValues: variables,
            operationName,
            fieldResolver: config.fieldResolver,
        };
        const executionDidEnd = extensionStack.executionDidStart({
            executionArgs,
        });
        try {
            /// ADDED CODE ///
            if (query && query[0] === '{') {
                return groupResolve_1.default({ queryString: query }, graphql.execute, executionArgs);
            }
            /////////////////
            return graphql.execute(executionArgs);
        }
        finally {
            executionDidEnd();
        }
    }
    async function sendResponse(response) {
        // We override errors, data, and extensions with the passed in response,
        // but keep other properties (like http)
        requestContext.response = extensionStack.willSendResponse({
            graphqlResponse: {
                ...requestContext.response,
                errors: response.errors,
                data: response.data,
                extensions: response.extensions,
            },
            context: requestContext.context,
        }).graphqlResponse;
        await dispatcher.invokeHookAsync('willSendResponse', requestContext);
        return requestContext.response;
    }
    function initializeRequestListenerDispatcher() {
        const requestListeners = [];
        if (config.plugins) {
            for (const plugin of config.plugins) {
                if (!plugin.requestDidStart)
                    continue;
                const listener = plugin.requestDidStart(requestContext);
                if (listener) {
                    requestListeners.push(listener);
                }
            }
        }
        return new dispatcher_1.Dispatcher(requestListeners);
    }
    function initializeExtensionStack() {
        graphql_extensions_1.enableGraphQLExtensions(config.schema);
        // If custom extension factories were provided, create per-request extension
        // objects.
        const extensions = config.extensions ? config.extensions.map(f => f()) : [];
        if (config.tracing) {
            extensions.push(new apollo_tracing_1.TracingExtension());
        }
        if (config.cacheControl) {
            cacheControlExtension = new apollo_cache_control_1.CacheControlExtension(config.cacheControl);
            extensions.push(cacheControlExtension);
        }
        return new graphql_extensions_1.GraphQLExtensionStack(extensions);
    }
    function initializeDataSources() {
        if (config.dataSources) {
            const context = requestContext.context;
            const dataSources = config.dataSources();
            for (const dataSource of Object.values(dataSources)) {
                if (dataSource.initialize) {
                    dataSource.initialize({
                        context,
                        cache: requestContext.cache,
                    });
                }
            }
            if ('dataSources' in context) {
                throw new Error('Please use the dataSources config option instead of putting dataSources on the context yourself.');
            }
            context.dataSources = dataSources;
        }
    }
}
exports.processGraphQLRequestNew = processGraphQLRequestNew;
//# sourceMappingURL=requestPipeline.js.map