"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
// import util from 'util';
const axios_1 = __importDefault(require("axios"));
const resolverDeets_1 = require("./resolverDeets");
const json_to_graphql_query_1 = require("json-to-graphql-query");
const asyncForEach = async (array, callback) => {
    for (let index = 0; index < array.length; index++) {
        await callback(array[index], index, array);
    }
};
exports.getPrimaryKey = (store, resolver) => {
    // TODO: check if primary key exists (store the schema on startup and reference it)
    const parent = store.requestGroup[resolver].parent;
    if (parent) {
        // if resolver has a parent - return parents name + 'Id'
        return (resolverDeets_1.resolverMap[parent].returnType === 'array' ? parent.substring(0, parent.length - 1) : parent) + 'Id';
    }
    else {
        // otherwise return default primary
        return 'id';
    }
};
exports.default = async (store) => {
    /**
     * Create a query object for a resolver
     */
    const createResolverQuery = (resolver) => {
        const parentResolver = store.requestGroup[resolver].parent;
        const res = {};
        // add fields
        store.requestGroup[resolver].fields.forEach((field) => {
            res[field] = true;
        });
        // if arguments are provided - add to the resolver
        if (Object.keys(store.requestGroup[resolver].arguments).length) {
            res.__args = {};
            Object.keys(store.requestGroup[resolver].arguments).forEach((key) => {
                res.__args[key] = store.requestGroup[resolver].arguments[key];
            });
        }
        // add primary key for all requests
        res[exports.getPrimaryKey(store, resolver)] = true;
        // if resolver has a parent of type array - add args
        if (parentResolver && store.keys[parentResolver]) {
            if (!res.__args)
                res.__args = {};
            // args[store.resolverRelationships[parentResolver][resolver]] = store.keys[parentResolver];
            res.__args[resolverDeets_1.resolverMap[parentResolver].nameAsSingular + 'Ids'] = store.keys[parentResolver];
        }
        return res;
    };
    /**
     * Sends a request off to a service
     */
    const sendRequest = async (resolver) => {
        // console.log('sendRequest called - resolver: ', resolver);
        // define the query object that will be sent to the service
        const query = {};
        query[resolver] = createResolverQuery(resolver);
        const resolversSent = new Set();
        resolversSent.add(resolver);
        // check if any other resolvers can be completed by the same service
        // TODO: this can be made more efficient if this is checked before this function is called
        // check which resolvers can be resolved 1st etc
        Object.keys(store.requestGroup).forEach((key) => {
            if (key !== resolver && // not the resolver called initially
                resolverDeets_1.resolverMap[key].serviceUrl === resolverDeets_1.resolverMap[resolver].serviceUrl // has the same resolving service
            ) {
                // console.log(key, 'is from the same service and can be resolved with', resolver);
                // same service, send as part of same request
                query[key] = createResolverQuery(key);
                resolversSent.add(key);
            }
        });
        // console.log('query pre organising', query);
        let resolversToBeSorted = Object.keys(query);
        while (resolversToBeSorted.length) {
            Object.keys(query).forEach((key) => {
                const parentResolver = store.requestGroup[key].parent;
                // if resolver has already been sorted
                if (!resolversToBeSorted.includes(key)) {
                    return;
                }
                // has a parent resolver - but parent has been resolved already (sent to another service for example)
                if (parentResolver && !store.resolversToSendRequestsTo.includes(parentResolver)) {
                    resolversToBeSorted = resolversToBeSorted.filter((e) => e !== key);
                    return;
                }
                const children = store.requestGroup[key].children;
                const childrenInResolversToBeSortedArray = children.some(r => resolversToBeSorted.indexOf(r) >= 0);
                // if a resolver has no children
                if (parentResolver && children.length === 0) {
                    query[parentResolver][key] = query[key];
                    delete query[key];
                    resolversToBeSorted = resolversToBeSorted.filter((e) => e !== key);
                    store.resolversToSendRequestsTo = store.resolversToSendRequestsTo.filter((e) => e !== key);
                    resolversSent.delete(key);
                    return;
                }
                // if a resovler has children + the children are NOT in the resolversToBeSorted array
                if (parentResolver && children.length > 0 && !childrenInResolversToBeSortedArray) {
                    query[parentResolver][key] = query[key];
                    delete query[key];
                    resolversToBeSorted = resolversToBeSorted.filter((e) => e !== key);
                    store.resolversToSendRequestsTo = store.resolversToSendRequestsTo.filter((e) => e !== key);
                    resolversSent.delete(key);
                    return;
                }
                // if a resovler has children + the children are in the resolversToBeSorted array
                if (parentResolver && children.length > 0 && childrenInResolversToBeSortedArray) {
                    return;
                }
                resolversToBeSorted = resolversToBeSorted.filter((e) => e !== key);
            });
        }
        // console.log('query:', query);
        // console.log("graphql'd query: ", jsonToGraphQLQuery(query, { pretty: true }));
        const res = await axios_1.default({
            url: resolverDeets_1.resolverMap[resolver].serviceUrl,
            method: 'post',
            data: {
                query: json_to_graphql_query_1.jsonToGraphQLQuery({ query }, { pretty: true }),
            },
        });
        // console.log('axios response: ', util.inspect(res.data, false, null));
        // console.log('resolversSent: ', resolversSent);
        resolversSent.forEach((key) => {
            if (Array.isArray(res.data.data[key])) {
                store.keys[key] = res.data.data[key].map((item) => item[exports.getPrimaryKey(store, key)]);
            }
            else {
                store.keys[key] = [];
                store.keys[key].push(res.data.data[key][exports.getPrimaryKey(store, key)]);
            }
            // add result to store
            if (!store.results[key]) {
                store.results[key] = {};
            }
            store.results[key].result = res.data.data[key];
            // remove resolver from resolversToSendRequestsTo
            store.resolversToSendRequestsTo = store.resolversToSendRequestsTo.filter((e) => e !== key);
        });
        // console.log('keys:', store.keys);
        // console.log('resolversToSendRequestsTo: ', store.resolversToSendRequestsTo);
        return res.data;
    };
    // create an array of services that are awaiting requests to be sent to them
    store.resolversToSendRequestsTo = Object.keys(store.requestGroup);
    // loop through resolvers until all have been completed
    while (store.resolversToSendRequestsTo.length) {
        await asyncForEach(store.resolversToSendRequestsTo, async (resolver) => {
            // skip if the resolver has been finished already in a grouped service call
            if (!store.resolversToSendRequestsTo.includes(resolver)) {
                return;
            }
            const parentResolver = store.requestGroup[resolver].parent;
            // check if a parent call is required 1st
            // if the parent resolver is still in the pending calls, skip this until next iteration, then try again
            if (parentResolver && store.resolversToSendRequestsTo.includes(parentResolver)) {
                // do nothing
            }
            else {
                await sendRequest(resolver);
            }
        });
    }
};
//# sourceMappingURL=sendRequestsToServices.js.map